\section{Basics}

> {-# LANGUAGE DeriveFunctor, DeriveFoldable, DeriveTraversable #-}

> module Basics
>     (  Fwd(..)
>     ,  Bwd(..)
>     ,  (<+>)
>     ,  (<>>)
>     ) where

> import Data.Foldable
> import Data.Monoid
> import Data.Traversable

We define our own types of forward (|Fwd|) and backward (|Bwd|) lists,
which are traversable functors and monoids. We always think of
contexts as being backwards lists and context suffixes (extensions) as
being forwards lists, and reflecting this in the code makes it clearer
what is going on.

> data Fwd a = F0 | a :> Fwd a
>     deriving (Eq, Show, Functor, Foldable, Traversable)

> data Bwd a = B0 | Bwd a :< a
>     deriving (Eq, Show, Functor, Foldable, Traversable)

> infixr 8 :>
> infixl 8 :<

> instance Monoid (Bwd a) where
>     mempty = B0
>     xs `mappend` B0         = xs
>     xs `mappend` (ys :< y)  = (xs `mappend` ys) :< y

> instance Monoid (Fwd a) where
>     mempty = F0
>     F0         `mappend` ys = ys
>     (x :> xs)  `mappend` ys = x :> (xs `mappend` ys)

> (<+>) :: Monoid a => a -> a -> a
> (<+>) = mappend

> (<>>) :: Bwd a -> Fwd a -> Fwd a
> infixl 8 <>>
> B0 <>> ys         = ys
> (xs :< x) <>> ys  = xs <>> (x :> ys)
