\section{GroupUnify}

> module GroupUnify (unifyGExp, embedVar) where

> import Basics
> import Group
> import Context


To unify a group expression, we negate one side and unify with the
group identity.

> unifyGExp :: Dimension -> Dimension -> Contextual ()
> unifyGExp d e = unifyId Nothing $ d -~ e

First we check if |e| is already the identity (in which case we win)
or is another constant (in which case we lose). Otherwise, we search
the context for group variables that occur in |e|. When we find one,
we either substitute it into the expression (if it has a definition)
or examine the coefficients to determine how to proceed. If its
coefficient |n| divides all the others, we can define it to solve the
equation. Otherwise, we either reduce the coefficients modulo |n| or
just collect the variable in the box and move it back in the context.

> unifyId :: Maybe TyName -> Dimension -> Contextual ()
> unifyId _Psi e
>   | isIdentity e  = return ()
>   | isConstant e  = fail "Unit mismatch!"
>   | otherwise     = onTopGE $
>     \ (alpha := d) ->
>     case lookupVariable alpha e of
>         Nothing  -> unifyId _Psi e >> restore
>         Just n   -> case d of
>             Just x   -> do
>                           modifyContext (<><| _Psi)
>                           unifyId Nothing (substGExp (alpha, n) x e)
>                           restore
>             Nothing  | n `dividesCoeffs` e -> do
>                           modifyContext (<><| _Psi)
>                           replace $ (alpha := GE (Just (pivot (alpha, n) e))) :> F0
>                      | (alpha, n) `notMaxCoeff` e -> do
>                           modifyContext (<><| _Psi)
>                           (p, beta) <- insertFreshVar $ pivot (alpha, n) e
>                           unifyId (Just beta) $ substGExp (alpha, n) p e
>                           replace $ (alpha := GE (Just p)) :> F0
>                      | numVariables e > 1 -> do
>                           demandNothing _Psi
>                           unifyId (Just alpha) e
>                           replace F0
>                      | otherwise -> fail "No way!"
>   where
>     _Gamma <><| Nothing     = _Gamma
>     _Gamma <><| Just alpha  = _Gamma :< A (alpha := GE Nothing)
>
>     demandNothing Nothing   = return ()
>     demandNothing (Just _)  = error "demandNothing: invariants violated!"


We can insert a fresh variable into a unit thus:

> insertFreshVar :: Dimension -> Contextual (Dimension, TyName)
> insertFreshVar d = do
>     beta <- freshName
>     return (d +~ embedVar beta, beta)
