\section{Inference}

> module Inference (infer) where

> import Basics
> import Context
> import Unify


> specialise :: Scheme -> Contextual Type
> specialise (Type tau) = return tau
> specialise (Bind d sigma) = do
>     beta <- fresh d
>     specialise (fmap (fromS beta) sigma)
>   where
>     fromS :: Integer -> Index Integer -> Integer
>     fromS beta  Z          = beta
>     fromS _     (S alpha)  = alpha

> (>=>) :: Suffix -> Type -> Scheme
> F0                     >=> tau = Type tau
> ((alpha := d) :> _Xi)  >=> tau = Bind d (fmap help (_Xi >=> tau))
>   where
>     help :: TyName -> Index TyName
>     help beta  | alpha == beta  = Z
>                | otherwise      = S beta


> generaliseOver ::  Contextual Type -> Contextual Scheme
> generaliseOver t = do
>     modifyContext (:< LetGoal)
>     tau  <- t
>     _Xi  <- skimContext F0
>     return (_Xi >=> tau)
>   where
>     skimContext :: Suffix -> Contextual Suffix
>     skimContext _Xi = do
>         _Gamma :< e <- getContext
>         putContext _Gamma
>         case e of
>             LetGoal   -> return _Xi
>             A alphaD  -> skimContext (alphaD :> _Xi)
>             T _       -> error "Unexpected term variable!"




> find :: TmName -> Contextual Scheme
> find x = getContext >>= help
>   where
>     help :: Context -> Contextual Scheme
>     help (_Gamma :< T (y ::: sigma)) | x == y  = return sigma
>     help (_Gamma :< _)                         = help _Gamma
>     help B0                                    = fail "Missing var!"



> (>-) :: TmEntry -> Contextual a -> Contextual a
> x ::: sigma >- ma = do
>     modifyContext (:< T (x ::: sigma))
>     a <- ma
>     modifyContext extract
>     return a 
>   where          
>     extract ::  Context -> Context
>     extract (_Gamma :< T (y ::: _)) | x == y = _Gamma
>     extract (_Gamma :< A alphaD) = (extract _Gamma) :< A alphaD
>     extract (_Gamma :< _)  = error "Bad context entry!"
>     extract B0             = error "Missing TM variable!"


> infer :: Term -> Contextual Type
>
> infer (X x) = find x >>= specialise
>
> infer (Lam x w) = do
>     alpha    <- fresh (TY Nothing)
>     upsilon  <- x ::: Type (V alpha) >- infer w
>     return (V alpha :-> upsilon)
>
> infer (f :$ a) = do
>     chi      <- infer f
>     upsilon  <- infer a
>     beta     <- fresh (TY Nothing)
>     unify chi (upsilon :-> V beta)
>     return (V beta)
>
> infer (Let x s w) = do
>     sigma <- generaliseOver (infer s)
>     x ::: sigma >- infer w
